import numpy as np
import xarray as xr
from warnings import warn
from .CDF import CDF
from .interpolate_along_dimension import interpolate_along_dimension
from .COAsT import COAsT

class CRPS():
    '''
    Object for handling and storing necessary information, methods and outputs
    for calculation of Continuous Ranked Probability Score. The object is
    initialized by passing it COAsT variables of model data, model domain and
    and observation object. CRPS can then be calculated using the
    CRPS.calculate() function. This will return an array of CRPS values
    (if desired) or will store them inside the object. They can be accessed
    from the object by calling CRPS.crps or CRPS['crps']

    Example basic usage::

        $ crps_obj = coast.CRPS(nemo_data, nemo_domain, altimetry)
        $ crps_list = crps_obj.crps # Get crps values
        $ crps.map_plot() # Plots CRPS on map
    '''
    
###############################################################################
#######                       ~Initialisation~                          #######
###############################################################################
    
    def __init__(self, model: COAsT, observations: COAsT,
                 var_name_mod:str, var_name_obs:str, nh_radius: float=20, 
                 nh_type: str="radius", cdf_type: str="empirical", 
                 time_interp: str="nearest"):
        """Initialisation of CRPS object.

        Args:
            mod_data (COAsT): COAsT model data object.
            mod_dom  (COAsT): COAsT model domain object.
            obs_object (OBSERVATION): COAsT OBSERVATION object.
            mod_var (str): Name of variable to use from model object.
            obs_var (str): Name of variable to use from observation object.
            nh_radius(float): Neighbourhood radius.
            nh_type(str): Neighbourhood type, either 'radius' or 'box'
            cdf_type(str): Type of CDF to use for model data.
                           Either 'empirical' or 'theoretical'
            time_interp(str): Type of time interpolation.
                              Either 'nearest' or 'linear'

        Returns:
            CRPS: Returns a new instance of a CRPS object.
        """
        #Input variables
        self.dataset      = observations[['longitude','latitude','time']]
        self.model        = model.dataset
        self.observations = observations.dataset
        self.nh_radius = nh_radius
        self.nh_type = nh_type
        self.cdf_type = cdf_type
        self.time_interp = time_interp
        self.var_name_mod = var_name_mod
        self.var_name_obs = var_name_obs
        self.dataset.attrs = {'title':'Continous Ranked Probability Score ' +
                              'for model('+ var_name_mod + 
                              ') vs observations(' + var_name_obs + '): '}
        # Output variables
        self.calculate()
        return
    
###############################################################################
#######                       ~General Routines~                        #######
###############################################################################
    
    def __getitem__(self, varstr:str):
        """Gets variable from __dict__"""
        return self.__dict__[varstr]

    def calculate(self):
        """Calculate CRPS values for specified variables/methods/radii."""
        tmp = self.calculate_sonf(self.model[self.var_name_mod], 
                                  self.observations[self.var_name_obs], 
                                  self.nh_radius, self.nh_type,
                                  self.cdf_type, self.time_interp)
        self.dataset['crps'] = tmp[0]
        self.dataset['n_model_pts']= tmp[1]
        self.dataset['contains_land'] = tmp[2]
        self.dataset['mean'] = np.nanmean(tmp[0])
        self.dataset['mean_noland'] = np.nanmean(tmp[0][tmp[2]==0])
        return 
    
    def calculate_sonf(self, model_data, obs_data, nh_radius: float, 
                       nh_type: str, cdf_type:str, time_interp:str):
        """Calculatues the Continuous Ranked Probability Score (CRPS)

        Calculatues the Continuous Ranked Probability Score (CRPS) using
        a single-observation and neighbourhood forecast (SONF). The statistic
        uses a comparison between the probability distributions of a model 
        neighbourhood subset and a single observation. The CRPS is calculated 
        independently for each observation. 
        """
        # Get relevant data and rename time dimension for interpolation
        model_data = model_data.rename({'t_dim':'time'})
        
        # Extract only x_dim, y_dim and t_dim dimensions. 
        # In case of other dimensions (e.g. depth), take first index
        for dim in model_data.dims:
            if dim not in ['x_dim', 'y_dim', 'time']:
                model_data = model_data.isel(dim=0)
    
        # Define output arrays
        n_neighbourhoods = obs_data.shape[0] 
        crps_list     = np.zeros( n_neighbourhoods )*np.nan
        n_model_pts   = np.zeros( n_neighbourhoods )*np.nan
        contains_land = np.zeros( n_neighbourhoods , dtype=bool)
        mod_cdf = None
        obs_cdf = None

        # Loop over neighbourhoods
        neighbourhood_indices = np.arange(0,n_neighbourhoods)
        for ii in neighbourhood_indices:
            
            print("\r Progress: [[ "+str(round(ii/n_neighbourhoods*100,2)) + 
                  '% ]]', end=" ", flush=True)
            
            # Neighbourhood centre
            cntr_lon = obs_data.longitude[ii]
            cntr_lat = obs_data.latitude[ii]
        
            # Get model neighbourhood subset using specified method
            if nh_type == "radius":
                subset_ind = self.subset_indices_by_distance(model_data.longitude,
                                  model_data.latitude, cntr_lon, cntr_lat, 
                                  nh_radius)
            elif nh_type == "box":
                raise NotImplementedError
            
            # Check that the model neighbourhood contains points
            if subset_ind[0].shape[0] == 0 or subset_ind[1].shape[0] == 0:
                crps_list[ii] = np.nan
            else:
                # Subset model data in time and space: model -> obs
                mod_subset = model_data.isel(y_dim = subset_ind[0],
                                               x_dim = subset_ind[1])
                mod_subset = mod_subset.interp(time = obs_data['time'][ii],
                                                   method = time_interp,
                                                   kwargs={'fill_value':'extrapolate'})
                
                #Check if neighbourhood contains a land value (TODO:mask)
                if any(np.isnan(mod_subset)):
                    contains_land[ii] = True
                # Check that neighbourhood contains a value
                if all(np.isnan(mod_subset)):
                    pass
                else:
                    # Create model and observation CDF objects
                    mod_cdf = CDF(mod_subset, cdf_type = cdf_type)
                    obs_cdf = CDF([obs_data[ii]], cdf_type = 'empirical')
                
                    # Calculate CRPS and put into output array
                    crps_list[ii] = mod_cdf.crps_fast(obs_data[ii])
                    n_model_pts[ii] = int(mod_subset.shape[0])
                    
        print("\r Complete.                             \n", end=" ", flush=True)

        return crps_list, n_model_pts, contains_land, mod_cdf, obs_cdf
    
    def subset_indices_by_distance(self, longitude, latitude,
                                   centre_lon: float, centre_lat: float, 
                                   radius: float):
        """
        This method returns a `tuple` of indices within the `radius` of the lon/lat point given by the user.

        Distance is calculated as haversine - see `self.calculate_haversine_distance`

        :param centre_lon: The longitude of the users central point
        :param centre_lat: The latitude of the users central point
        :param radius: The haversine distance (in km) from the central point
        :return: All indices in a `tuple` with the haversine distance of the central point
        """

        # Calculate the distances between every model point and the specified
        # centre. Calls another routine dist_haversine.

        dist = self.calculate_haversine_distance(centre_lon, centre_lat, 
                                                 longitude, latitude)
        indices_bool = dist < radius
        indices = np.where(indices_bool.compute())

        return xr.DataArray(indices[0]), xr.DataArray(indices[1])
    
    def calculate_haversine_distance(self, lon1, lat1, lon2, lat2):
        '''
        # Estimation of geographical distance using the Haversine function.
        # Input can be single values or 1D arrays of locations. This
        # does NOT create a distance matrix but outputs another 1D array.
        # This works for either location vectors of equal length OR a single loc
        # and an arbitrary length location vector.
        #
        # lon1, lat1 :: Location(s) 1.
        # lon2, lat2 :: Location(s) 2.
        '''

        # Convert to radians for calculations
        lon1 = xr.ufuncs.deg2rad(lon1)
        lat1 = xr.ufuncs.deg2rad(lat1)
        lon2 = xr.ufuncs.deg2rad(lon2)
        lat2 = xr.ufuncs.deg2rad(lat2)

        # Latitude and longitude differences
        dlat = (lat2 - lat1) / 2
        dlon = (lon2 - lon1) / 2

        # Haversine function.
        distance = xr.ufuncs.sin(dlat) ** 2 + xr.ufuncs.cos(lat1) * xr.ufuncs.cos(lat2) * xr.ufuncs.sin(dlon) ** 2
        distance = 2 * 6371.007176 * xr.ufuncs.arcsin(xr.ufuncs.sqrt(distance))

        return distance
    
    
###############################################################################
#######                       ~PLOTTING ROUTINES~                       #######
###############################################################################
    
    def cdf_plot(self, index):
        """A comparison plot of the model and observation CDFs.

        Args:
            index (int): Observation index to plot CDFs for (single index).

        Returns:
            Figure and axes objects for the resulting image.
        """
        index=[index]
        tmp = self.calculate_sonf(self.model[self.var_name_mod], 
                                  self.observations[self.var_name_obs][index],
                                  self.nh_radius, self.nh_type, self.cdf_type,
                                  self.time_interp)
        crps_tmp = tmp[0]
        n_mod_pts = tmp[1]
        contains_land = tmp[2]
        mod_cdf = tmp[3]
        obs_cdf = tmp[4]
        fig, ax = mod_cdf.diff_plot(obs_cdf)
        titlestr = 'CRPS = ' + str(round( crps_tmp[0], 3)) + '\n'
        titlestr = titlestr + '# Model Points : ' + str(n_mod_pts[0]) + '  |  '
        titlestr = titlestr + 'Contains land : ' + str(bool(contains_land[0]))
        ax.set_title(titlestr)
        ax.grid()
        ax.legend(['Model', 'Observations'])
        return fig, ax

    def map_plot(self, stats_var: str='crps'):
        """Plots CRPS (or other variables) at observation locations on a map

        Args:
            stats_var (str): Either 'crps', 'n_model_pts' or 'contains_land'.

        Returns:
            Figure and axes objects for the resulting image.
        """
        try:
            import cartopy.crs as ccrs  # mapping plots
            import cartopy.feature  # add rivers, regional boundaries etc
            from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER  # deg symb
            from cartopy.feature import NaturalEarthFeature  # fine resolution coastline
            import matplotlib.pyplot as plt
        except ImportError:
            import sys
            warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
            sys.exit(-1)

        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca()
        ax = plt.subplot(1, 1, 1, projection=ccrs.PlateCarree())
        plt.scatter(self.dataset.longitude, self.dataset.latitude, 
                    c=self.dataset[stats_var])
        plt.colorbar()
        ax.add_feature(cartopy.feature.BORDERS, linestyle=':')
        coast = NaturalEarthFeature(category='physical', scale='50m',
                                    facecolor='none', name='coastline')
        ax.add_feature(coast, edgecolor='gray')
        gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                          linewidth=0.5, color='gray', alpha=0.5, linestyle='-')
        gl.top_labels = False
        gl.bottom_labels = True
        gl.right_labels = False
        gl.left_labels = True
        gl.xformatter = LONGITUDE_FORMATTER
        gl.yformatter = LATITUDE_FORMATTER
        title_dict = {'crps': 'Continuous Rank Probability Score',
                      'n_model_pts' : 'Number of Model Points Used',
                      'contains_land': 'CRPS Values that contain land'}
        try:
            plt.title(title_dict[stats_var])
        except:
            pass

        plt.show()
        return fig, ax



from .COAsT import COAsT
from .OBSERVATION import OBSERVATION
from warnings import warn
import numpy as np
import xarray as xr

class ALTIMETRY(OBSERVATION):
    '''
    An object for reading, storing and manipulating altimetry data.
    Currently the objecgt is set up for reading altimetry netCDF data from
    the CMEMS database.
    
    Data should be stored in an xarray.Dataset, in the form:
        
    * Date Format Overview *
        
        1. A single dimension (time).
        2. Three coordinates: time, latitude, longitude. All lie on the time
           dimension.
        3. Observed variable DataArrays on the time dimension.
        
    There are currently no naming conventions for the variables however
    examples from the CMEMS database include sla_filtered, sla_unfiltered and
    mdt (mean dynamic topography).
    
    * Methods Overview *
    
    1. __init__(): Initialises an ALTIMETRY object.
    2. quick_plot(): Makes a quick plot of the data inside the object. 
    3. obs_operator(): For interpolating model data to this object.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.observation_type = 'moving'
        self.dataset = self.dataset.rename_dims(self.dim_mapping)
        #self.dataset.attrs = {}
        return

    def set_dimension_mapping(self):
        self.dim_mapping = {'time':'t_dim'}

    def set_variable_mapping(self):
        self.var_mapping = None

    def quick_plot(self, var: str=None):
        '''
        Quick geographical plot of altimetry data for a specified variable
    
        Example usage:
        --------------
        # Have a quick look at sla_filtered
        altimetry.quick_plot('sla_filtered')

        Parameters
        ----------
        var (str) : Variable to plot. Default is None, in which case only
            locations are plotted.
        Returns
        -------
        Figure and axes objects
        '''
        try:
            import cartopy.crs as ccrs  # mapping plots
            import cartopy.feature  # add rivers, regional boundaries etc
            from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER  # deg symb
            from cartopy.feature import NaturalEarthFeature  # fine resolution coastline
        except ImportError:
            import sys
            warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
            sys.exit(-1)

        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca()
        ax = plt.subplot(1, 1, 1, projection=ccrs.PlateCarree())

        if var is None:
            cset = self.dataset.plot.scatter(x='longitude',y='latitude')
        else:
            cset = self.dataset.plot.scatter(x='longitude',y='latitude',hue=var)

        ax.add_feature(cartopy.feature.BORDERS, linestyle=':')
        coast = NaturalEarthFeature(category='physical', scale='50m',
                                    facecolor=[0.8,0.8,0.8], name='coastline',
                                    alpha=0.5)
        ax.add_feature(coast, edgecolor='gray')

        gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                          linewidth=0.5, color='gray', alpha=0.5, linestyle='-')

        gl.top_labels = False
        gl.bottom_labels = True
        gl.right_labels = False
        gl.left_labels = True
        gl.xformatter = LONGITUDE_FORMATTER
        gl.yformatter = LATITUDE_FORMATTER

        plt.show()
        return fig, ax
    
    def obs_operator(self, model, mod_var_name:str, 
                                time_interp = 'nearest'):
        '''
        For interpolating a model dataarray onto altimetry locations and times.
        
        For ALTIMETRY, the interpolation is done independently in two steps:
            1. Horizontal space
            2. Time
        Model data is taken at the surface if necessary (0 index). 
    
        Example usage:
        --------------
        altimetry.obs_operator(nemo_obj, 'sossheig')

        Parameters
        ----------
        model : model object (e.g. NEMO)
        mod_var: variable name string to use from model object
        time_interp: time interpolation method (optional, default: 'nearest')
            This can take any string scipy.interpolate would take. e.g.
            'nearest', 'linear' or 'cubic'
        Returns
        -------
        Adds a DataArray to self.dataset, containing interpolated values.
        '''
        
        # Get data arrays
        mod_var = model.dataset[mod_var_name]
        
        # Depth interpolation -> for now just take 0 index
        if 'z_dim' in mod_var.dims:
            mod_var = mod_var.isel(z_dim=0).squeeze()
        
        # Cast lat/lon to numpy arrays
        obs_lon = np.array(self.dataset.longitude).flatten()
        obs_lat = np.array(self.dataset.latitude).flatten()
        
        interpolated = model.interpolate_in_space(mod_var, obs_lon, 
                                                        obs_lat)
        
        # Interpolate in time if t_dim exists in model array
        if 't_dim' in mod_var.dims:
            interpolated = model.interpolate_in_time(interpolated, 
                                                     self.dataset.time,
                                                     interp_method=time_interp)
            # Take diagonal from interpolated array (which contains too many points)
            diag_len = interpolated.shape[0]
            diag_ind = xr.DataArray(np.arange(0, diag_len))
            interpolated = interpolated.isel(interp_dim=diag_ind, t_dim=diag_ind)
            interpolated = interpolated.swap_dims({'dim_0':'t_dim'})

        # Store interpolated array in dataset
        new_var_name = 'interp_' + mod_var_name
        self.dataset[new_var_name] = interpolated
        return



from setuptools import setup
from sys import argv
from types import SimpleNamespace


PACKAGE = SimpleNamespace(**{
    "name": "COAsT",
    "version": "0.2.1a22",
    "description": "This is the Coast Ocean Assessment Tool",
    "url": "https://www.bodc.ac.uk",
    "download_url": "https://github.com/British-Oceanographic-Data-Centre/COAsT/",
    "author": "British Oceanographic Data Centre (BODC)",
    "author_email": "bodcsoft@bodc.ac.uk",
    "license": "Put something here",
    "license_family": "OTHER",
    "setup_requires": [
        "wheel"
    ],
    "classifiers": [
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Hydrology",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7"
    ],
    "keywords": [
        "NEMO",
        "shallow water",
        "ocean assessment"
    ],
    "project_urls": {
        "documentation": "https://british-oceanographic-data-centre.github.io/COAsT/"
    },
    "install_requires": [
        "dask[complete]",
        "xarray",
        "numpy",
        "matplotlib",
        "netCDF4",
        "scipy",
        "gsw"
    ],
    "python_requires": ">=3",
    "packages": [
        "coast"
    ],
    "include_package_data": True,
    "github": "British-Oceanographic-Data-Centre"
})


def generate_conda(directory="conda"):
    import oyaml as yaml  # OYaml is used to preserve the order of the metadata dict in YAML output
    from collections import OrderedDict
    from os import path

    package_metadata = OrderedDict({
        "package": {
            "name": PACKAGE.name.lower(),
            "version": PACKAGE.version
        },
        "source": {
            "url": f"https://pypi.io/packages/source/{PACKAGE.name[0]}/{PACKAGE.name}/{PACKAGE.name}-{PACKAGE.version}.tar.gz"
        },
        "requirements": {
            "host": PACKAGE.install_requires,
            "run": PACKAGE.install_requires
        },
        "test": {
            "imports": [
                "coast"
            ]
        },
        "about": {
            "home": PACKAGE.url,
            "license": PACKAGE.license,
            "license_family": PACKAGE.license_family,
            "license_file": "",
            "summary": PACKAGE.description,
            "doc_url": "",
            "dev_url": "",
        },
        "extra": {
            "recipe-maintainers": [
                PACKAGE.github
            ]
        }
    })

    yaml_path = path.join(directory, "meta.yaml")
    yaml.Dumper.ignore_aliases = lambda *args: True  # Dummy function to return true
    with open(yaml_path, "w") as meta_file:
        yaml.dump(package_metadata, meta_file)


if __name__ == "__main__":
    if argv[1] == "conda":
        generate_conda()
    else:
        setup(
            name=PACKAGE.name,
            version=PACKAGE.version,
            description=PACKAGE.description,
            url=PACKAGE.url,
            download_url=PACKAGE.download_url,
            author=PACKAGE.author,
            author_email=PACKAGE.author_email,
            license=PACKAGE.license,  # TODO,
            setup_requires=PACKAGE.setup_requires,
            classifiers=PACKAGE.classifiers,
            keywords=PACKAGE.keywords,
            project_urls=PACKAGE.project_urls,
            install_requires=PACKAGE.install_requires,
            python_requires=PACKAGE.python_requires,
            packages=PACKAGE.packages,
            include_package_data=PACKAGE.include_package_data
        )



import xarray as xr
import numpy as np
import matplotlib.pyplot as plt

class CDF():
    '''
    An object for storing Cumulative Distribution Function information.
    '''
    
    def __init__(self, sample, cdf_type: str='empirical', 
                 cdf_func: str='gaussian'):
        """Initialisation of CDF object.

        Args:
            sample (array): Data sample over which to estimate CDF
            cdf_type (str): Either 'empirical' or 'theoretical'.
            cdf_func (str): Function type if cdf_type='theoretical'. Presently
                            only 'gaussian'.

        Returns:
            New CDF object.
        """
        self.cdf_type = cdf_type
        self.cdf_func = cdf_func
        self.sample = sample
        self.sample_size = len(sample)
        self.mu       = np.nanmean(sample)
        self.sigma    = np.nanstd(sample)
        self._cdf_type_options = ['empirical', 'theoretical']
        self._cdf_func_options = ['gaussian']
        self.plot_xmin, self.plot_xmax = self.set_x_bounds()
        
    def set_x_bounds(self):
        ''' Calculate x bounds for CDF plotting '''
        # Is input a single value (st. dev == 0)
        single_value = True if self.sigma == 0 else False
        if single_value: 
            self.cdf_type = 'empirical'
        
        # Calculate x bounds as 5 std dev. either side of the mean
        if single_value:
            xmin = self.mu-1
            xmax = self.mu+1
        elif self.cdf_func == 'gaussian':
            xmin = self.mu - 5*self.sigma
            xmax = self.mu + 5*self.sigma
        return xmin, xmax
    
    def build_discrete_cdf(self, x: np.ndarray=None, n_pts: int=1000):
        """Builds a discrete CDF for plotting and direct comparison.

        Args:
            x (array): x-values over which to calculate discrete CDF values.
                       If none, these will be determined using mu and sigma.
            n_pts (int): n_pts to use for x if x=None.

        Returns:
            x and y arrays for discrete CDF.
        """
        
        # Build discrete X bounds
        if x is None:
            x = np.linspace(self.plot_xmin, self.plot_xmax, n_pts)
        
        if self.cdf_type == "empirical":
            y = self.empirical_distribution(x, self.sample)

        elif self.cdf_type == "theoretical": 
            if self.cdf_func == 'gaussian':
                y = self.cumulative_distribution(mu=self.mu, sigma=self.sigma,
                                                 x=x, cdf_func=self.cdf_func)
            else:
                raise NotImplementedError
        else:
            raise Exception('CDF Type must be empirical or theoretical')

        return x, y
    
    def normal_distribution(self, mu: float=0, sigma: float=1, 
                            x: np.ndarray=None, n_pts: int=1000):
        """Generates a discrete normal distribution.

        Keyword arguments:
        x     -- Arbitrary array of x-values
        mu    -- Distribution mean
        sigma -- Distribution standard deviation
        
        return: Array of len(x) containing the normal values calculated from
                the elements of x.
        """
        if x is None:
            x = np.linspace( mu-5*sigma, mu+5*sigma, n_pts)
        term1 = sigma*np.sqrt( 2*np.pi )
        term1 = 1/term1
        exponent = -0.5*((x-mu)/sigma)**2
        return term1*np.exp( exponent )

    def cumulative_distribution(self, mu: float=0, sigma: float=1, 
                                x: np.ndarray=None, cdf_func: str='gaussian'):
        """Integrates under a discrete PDF to obtain an estimated CDF.

        Keyword arguments:
        x   -- Arbitrary array of x-values
        pdf -- PDF corresponding to values in x. E.g. as generated using
               normal_distribution.
        
        return: Array of len(x) containing the discrete cumulative values 
                estimated using the integral under the provided PDF.
        """
        if cdf_func=='gaussian': #If Gaussian, integrate under pdf
            pdf = self.normal_distribution(mu=mu, sigma=sigma, x=x)
            cdf = [np.trapz(pdf[:ii],x[:ii]) for ii in range(0,len(x))]
        else: 
            raise NotImplementedError
        return np.array(cdf)
    
    def empirical_distribution(self, x, sample):
        """Estimates a CDF empirically.

        Keyword arguments:
        x      -- Array of x-values over which to generate distribution
        sample -- Sample to use to generate distribution
        
        return: Array of len(x) containing corresponding EDF values
        """
        sample = np.array(sample)
        sample = sample[~np.isnan(sample)]
        sample = np.sort(sample)
        edf = np.zeros(len(x))
        n_sample = len(sample)
        for ss in sample:
            edf[x>ss] = edf[x>ss] + 1/n_sample
        return xr.DataArray(edf)
    
    def crps(self, xa):
        """Calculate CRPS as outlined in Hersbach et al. 2000

        Args:
            xa (float): A single 'observation' value which to compare against
                        CDF.

        Returns:
            A single CRPS value.
        """
        
        def calc(alpha, beta, p):
            return alpha * p**2 + beta*(1 - p)**2
        crps_sum = 0
        sample = np.array(self.sample).flatten()
        sample = sample[~np.isnan(sample)]
        sample = np.sort(sample)
        sample_size = len(sample)
        
        # All intervals within range of the sample distribution
        for ii in range(0, sample_size-1):
            p = (ii+1)/sample_size
            if xa > sample[ii+1]:
                alpha = sample[ii+1] - sample[ii]
                beta = 0
            elif xa < sample[ii]:
                alpha = 0
                beta = sample[ii+1] - sample[ii]
            else:
                alpha = xa - sample[ii]
                beta = sample[ii+1] - xa
            crps = calc(alpha, beta, p)
            crps_sum = crps_sum + crps
        # Intervals 0 and N, where p = 0 and 1 respectively
        if xa < sample[0]:
            p=0
            alpha = 0
            beta = sample[0] - xa
            crps = calc(alpha, beta, p)
            crps_sum = crps_sum + crps
        elif xa > sample[-1]:
            p=1
            alpha = xa - sample[-1]
            beta = 0
            crps = calc(alpha, beta, p)
            crps_sum = crps_sum + crps
            
        return crps_sum
    
    def crps_fast(self, xa):
        """Faster version of crps method, using multiple numpy arrays.
           For large sample sizes, this is considerably faster but uses more
           memory.

        Args:
            xa (float): A single 'observation' value which to compare against
                        CDF.

        Returns:
            A single CRPS value.
        """
        
        def calc(alpha, beta, p):
            return alpha * p**2 + beta*(1 - p)**2
        xa = float(xa)
        crps_sum = 0
        sample = np.array(self.sample).flatten()
        sample = sample[~np.isnan(sample)]
        sample = np.sort(sample)
        sample_size = len(sample)
        
        alpha = np.zeros(sample_size-1)
        beta= np.zeros(sample_size-1)
        # sample[1:] = upper bounds, and vice versa
        
        tmp = sample[1:] - sample[:-1]
        tmp_logic = sample[1:]<xa
        alpha[tmp_logic] = tmp[tmp_logic]
                               
        tmp_logic = sample[:-1]>xa
        beta[tmp_logic] = tmp[tmp_logic]
        
        tmp_logic = ( sample[1:]>xa )*( sample[:-1]<xa )
        tmp = xa - sample[:-1]
        alpha[tmp_logic] = tmp[tmp_logic]
        tmp = sample[1:] - xa
        beta[tmp_logic] = tmp[tmp_logic]
        
        p = np.arange(1,sample_size)/sample_size
        c = alpha*p**2 + beta*(1-p)**2
        crps_sum = np.sum(c)
        
        # Intervals 0 and N, where p = 0 and 1 respectively
        if xa < sample[0]:
            p=0
            alpha = 0
            beta = sample[0] - xa
            crps = calc(alpha, beta, p)
            crps_sum = crps_sum + crps
        elif xa > sample[-1]:
            p=1
            alpha = xa - sample[-1]
            beta = 0
            crps = calc(alpha, beta, p)
            crps_sum = crps_sum + crps
            
        return crps_sum
    
    def get_common_x(self, other, n_pts=1000):
        """Generates a common x vector for two CDF objects."""
        xmin = min(self.plot_xmin, other.plot_xmin)
        xmax = max(self.plot_xmax, other.plot_xmax)
        common_x = np.linspace(xmin, xmax, n_pts)
        return common_x
    
    def quick_plot(self):
        """ A quick plot showing the CDF contained in this object."""
        ax = plt.subplot(111)
        x,y = self.build_discrete_cdf()
        ax.plot(x, y)
        ax.grid()
        return
    
    def diff_plot(self, other):
        """Plots two CDFS on one plot, with the difference shaded"""
        fig = plt.figure()
        ax = plt.subplot(111)
        x = self.get_common_x(other)
        dum, mod_y = self.build_discrete_cdf(x)
        ax.plot(x, mod_y, c='k', linestyle='--')
        dum, obs_y = other.build_discrete_cdf(x)
        ax.plot(x, obs_y, linestyle='--')
        ax.fill_between(x, mod_y, obs_y, alpha=0.5)
        return fig, ax


from .COAsT import COAsT
from .COAsT import setup_dask_clinet
from .DOMAIN import DOMAIN
from .NEMO import NEMO
from .TRANSECT import Transect
from .ALTIMETRY import ALTIMETRY
from .OBSERVATION import OBSERVATION
from .CDF import CDF
from .CRPS import CRPS
from .INTERNALTIDE import INTERNALTIDE
from .TIDEGAUGE import TIDEGAUGE



from .COAsT import COAsT
import xarray as xr
import numpy as np
# from dask import delayed, compute, visualize
# import graphviz
import matplotlib.pyplot as plt
import sklearn.neighbors as nb
import gsw
from scipy.interpolate import interp1d
from scipy.interpolate import griddata
import warnings

class NEMO(COAsT):
    """
    Words to describe the NEMO class

    kwargs -- define addition keyworded arguemts for domain file. E.g. ln_sco=1
    if using s-scoord in an old domain file that does not carry this flag.
    """
    def __init__(self, fn_data=None, fn_domain=None, grid_ref='t-grid',
                 chunks: dict=None, multiple=False,
                 workers=2, threads=2, memory_limit_per_worker='2GB', **kwargs):
        self.dataset = xr.Dataset()
        self.grid_ref = grid_ref.lower()
        self.domain_loaded = False

        self.set_dimension_mapping()
        self.set_variable_mapping()
        if fn_data is not None:
            self.load(fn_data, chunks, multiple)
        self.set_dimension_names(self.dim_mapping)
        self.set_variable_names(self.var_mapping)

        if fn_domain is None:
            self.filename_domain = "" # empty store for domain fileanme
            pass
            #print("No NEMO domain specified, only limited functionality"+
            #      " will be available")
        else:
            self.filename_domain = fn_domain # store domain fileanme
            dataset_domain = self.load_domain(fn_domain, chunks)

            # Define extra domain attributes using kwargs dictionary
            ## This is a bit of a placeholder. Some domain/nemo files will have missing variables
            for key,value in kwargs.items():
                dataset_domain[key] = value

            if fn_data is not None:
                dataset_domain = self.trim_domain_size( dataset_domain )
            self.set_timezero_depths(dataset_domain) # THIS ADDS TO dataset_domain. Should it be 'return'ed (as in trim_domain_size) or is implicit OK?
            self.merge_domain_into_dataset(dataset_domain)

    def set_dimension_mapping(self):
        self.dim_mapping = {'time_counter':'t_dim', 'deptht':'z_dim',
                            'depthu':'z_dim', 'depthv':'z_dim',
                            'y':'y_dim', 'x':'x_dim'}
        self.dim_mapping_domain = {'t':'t_dim0', 'x':'x_dim', 'y':'y_dim',
                                   'z':'z_dim'}


    def set_variable_mapping(self):
        # Variable names remapped  within NEMO object
        self.var_mapping = {'time_counter':'time',
                            'votemper' : 'temperature',
                            'thetao' : 'temperature',
                            'temp' : 'temperature',
                            'so' : 'salinity',
                            'vosaline' : 'salinity'}
        # Variable names mapped from domain to NEMO object
        # NAMES NOT SET IN STONE.
        self.var_mapping_domain = {'time_counter' : 'time0',
                                   'glamt':'longitude', 'glamu':'longitude',
                                   'glamv':'longitude','glamf':'longitude',
                                   'gphit':'latitude', 'gphiu':'latitude',
                                   'gphiv':'latitude', 'gphif':'latitude',
                                   'e1t':'e1', 'e1u':'e1',
                                   'e1v':'e1', 'e1f':'e1',
                                   'e2t':'e2', 'e2u':'e2',
                                   'e2v':'e2', 'e2f':'e2',
                                   'ff_t':'ff', 'ff_f':'ff',
                                   'e3t_0':'e3_0', 'e3w_0':'e3_0',
                                   'e3u_0':'e3_0', 'e3v_0':'e3_0',
                                   'e3f_0':'e3_0',
                                   'tmask':'mask',
                                   'depthf_0':'depth_0',
                                   'depthu_0':'depth_0', 'depthv_0':'depth_0',
                                   'depthw_0':'depth_0', 'deptht_0':'depth_0',
                                   'ln_sco':'ln_sco'}

    def load_domain(self, fn_domain, chunks):
        ''' Loads domain file and renames dimensions with dim_mapping_domain'''
        # Load xarrat dataset
        dataset_domain = xr.open_dataset(fn_domain)
        self.domain_loaded = True
        # Rename dimensions
        for key, value in self.dim_mapping_domain.items():
            try:
                dataset_domain = dataset_domain.rename_dims({ key : value })
            except:
                print('pass: {}: {}', key, value)
                pass

        return dataset_domain

    def merge_domain_into_dataset(self, dataset_domain):
        ''' Merge domain dataset variables into self.dataset, using grid_ref'''
        # Define grid independent variables to pull across
        not_grid_vars = ['jpiglo', 'jpjglo','jpkglo','jperio',
                         'ln_zco', 'ln_zps', 'ln_sco', 'ln_isfcav']

        # Define grid specific variables to pull across
        if self.grid_ref == 'u-grid':
            grid_vars = ['glamu', 'gphiu', 'e1u', 'e2u', 'e3u_0', 'depthu_0'] #What about e3vw
        elif self.grid_ref == 'v-grid':
            grid_vars = ['glamv', 'gphiv', 'e1v', 'e2v', 'e3v_0', 'depthv_0']
        elif self.grid_ref == 't-grid':
            grid_vars = ['glamt', 'gphit', 'e1t', 'e2t', 'e3t_0', 'deptht_0', 'tmask']
        elif self.grid_ref == 'w-grid':
            grid_vars = ['glamt', 'gphit', 'e1t', 'e2t', 'e3w_0', 'depthw_0']
        elif self.grid_ref == 'f-grid':
            grid_vars = ['glamf', 'gphif', 'e1f', 'e2f', 'e3f_0', 'depthf_0']

        all_vars = grid_vars + not_grid_vars

        # Trim domain DataArray area if necessary.
        self.copy_domain_vars_to_dataset( dataset_domain, grid_vars )

        # Reset & set specified coordinates
        coord_vars = ['longitude', 'latitude', 'time', 'depth_0']
        self.dataset = self.dataset.reset_coords()
        for var in coord_vars:
            try:
                self.dataset = self.dataset.set_coords(var)
            except:
                pass

        # Delete specified variables
        # MIGHT NEEWD TO DELETE OTHEHR DEPTH VARS ON OTHER GRIDS?
        delete_vars = ['nav_lat', 'nav_lon', 'deptht']
        for var in delete_vars:
            try:
                self.dataset = self.dataset.drop(var)
            except:
                pass


    def __getitem__(self, name: str):
        return self.dataset[name]

    def set_grid_ref_attr(self):
        self.grid_ref_attr_mapping = {'temperature' : 't-grid',
                                'coast_name_for_u_velocity' : 'u-grid',
                                'coast_name_for_v_velocity' : 'v-grid',
                                'coast_name_for_w_velocity' : 'w-grid',
                                'coast_name_for_vorticity'  : 'f-grid' }
        #self.grid_ref_attr_mapping = None

    def get_contour_complex(self, var, points_x, points_y, points_z, tolerance: int = 0.2):
        smaller = self.dataset[var].sel(z=points_z, x=points_x, y=points_y, method='nearest', tolerance=tolerance)
        return smaller

    def set_timezero_depths(self, dataset_domain):

        """
        Calculates the depths at time zero (from the domain_cfg input file)
        for the appropriate grid.

        The depths are assigned to domain_dataset.depth_0

        """

        try:
            if self.grid_ref == 't-grid':
                e3w_0 = np.squeeze( dataset_domain.e3w_0.values )
                depth_0 = np.zeros_like( e3w_0 )
                depth_0[0,:,:] = 0.5 * e3w_0[0,:,:]
                depth_0[1:,:,:] = depth_0[0,:,:] + np.cumsum( e3w_0[1:,:,:], axis=0 )
            elif self.grid_ref == 'w-grid':
                e3t_0 = np.squeeze( dataset_domain.e3t_0.values )
                depth_0 = np.zeros_like( e3t_0 )
                depth_0[0,:,:] = 0.0
                depth_0[1:,:,:] = np.cumsum( e3t_0, axis=0 )[:-1,:,:]
            elif self.grid_ref == 'u-grid':
                e3w_0 = dataset_domain.e3w_0.values.squeeze()
                e3w_0_on_u = 0.5 * ( e3w_0[:,:,:-1] + e3w_0[:,:,1:] )
                depth_0 = np.zeros_like( e3w_0 )
                depth_0[0,:,:-1] = 0.5 * e3w_0_on_u[0,:,:]
                depth_0[1:,:,:-1] = depth_0[0,:,:-1] + np.cumsum( e3w_0_on_u[1:,:,:], axis=0 )
            elif self.grid_ref == 'v-grid':
                e3w_0 = dataset_domain.e3w_0.values.squeeze()
                e3w_0_on_v = 0.5 * ( e3w_0[:,:-1,:] + e3w_0[:,1:,:] )
                depth_0 = np.zeros_like( e3w_0 )
                depth_0[0,:-1,:] = 0.5 * e3w_0_on_v[0,:,:]
                depth_0[1:,:-1,:] = depth_0[0,:-1,:] + np.cumsum( e3w_0_on_v[1:,:,:], axis=0 )
            elif self.grid_ref == 'f-grid':
                e3w_0 = dataset_domain.e3w_0.values.squeeze()
                e3w_0_on_f = 0.25 * ( e3w_0[:,:-1,:-1] + e3w_0[:,1:,:-1] +
                                     e3w_0[:,:-1,:-1] + e3w_0[:,:-1,1:] )
                depth_0 = np.zeros_like( e3w_0 )
                depth_0[0,:-1,:-1] = 0.5 * e3w_0_on_f[0,:,:]
                depth_0[1:,:-1,:-1] = depth_0[0,:-1,:-1] + np.cumsum( e3w_0_on_f[1:,:,:], axis=0 )
            else:
                raise ValueError(str(self) + ": " + self.grid_ref + " depth calculation not implemented")
            # Write the depth_0 variable to the domain_dataset DataSet, with grid type
            dataset_domain[f"depth{self.grid_ref.replace('-grid','')}_0"] = xr.DataArray(depth_0,
                    dims=['z_dim', 'y_dim', 'x_dim'],
                    attrs={'units':'m',
                    'standard_name': 'Depth at time zero on the {}'.format(self.grid_ref)})
        except ValueError as err:
            print(err)

        return

    # Add subset method to NEMO class
    def subset_indices(self, start: tuple, end: tuple) -> tuple:
        """
        based on transect_indices, this method looks to return all indices between the given points.
        This results in a 'box' (Quadrilateral) of indices.
        consequently the returned lists may have different lengths.
        :param start: A lat/lon pair
        :param end: A lat/lon pair
        :return: list of y indices, list of x indices,
        """

        [j1, i1] = self.find_j_i(start[0], start[1])  # lat , lon
        [j2, i2] = self.find_j_i(end[0], end[1])  # lat , lon

        return list(np.arange(j1, j2+1)), list(np.arange(i1, i2+1))

    def find_j_i(self, lat: float, lon: float):
        """
        A routine to find the nearest y x coordinates for a given latitude and longitude
        Usage: [y,x] = find_j_i(49, -12)

        :param lat: latitude
        :param lon: longitude
        :return: the y and x coordinates for the NEMO object's grid_ref, i.e. t,u,v,f,w.
        """

        dist2 = xr.ufuncs.square(self.dataset.latitude - lat) + xr.ufuncs.square(self.dataset.longitude - lon)
        [y, x] = np.unravel_index(dist2.argmin(), dist2.shape)
        return [y, x]

    def find_j_i_domain(self, lat: float, lon: float, dataset_domain: xr.DataArray):
        """
        A routine to find the nearest y x coordinates for a given latitude and longitude
        Usage: [y,x] = find_j_i(49, -12, dataset_domain)

        :param lat: latitude
        :param lon: longitude
        :param grid_ref: the gphi/glam version a user wishes to search over
        :return: the y and x coordinates for the given grid_ref variable within the domain file
        """

        internal_lat = dataset_domain[f"gphi{self.grid_ref.replace('-grid','')}"]
        internal_lon = dataset_domain[f"glam{self.grid_ref.replace('-grid','')}"]
        dist2 = xr.ufuncs.square(internal_lat - lat) \
                + xr.ufuncs.square(internal_lon - lon)
        [_, y, x] = np.unravel_index(dist2.argmin(), dist2.shape)
        return [y, x]

    def transect_indices(self, start: tuple, end: tuple) -> tuple:
        """
        This method returns the indices of a simple straight line transect between two
        lat lon points defined on the NEMO object's grid_ref, i.e. t,u,v,f,w.

        :type start: tuple A lat/lon pair
        :type end: tuple A lat/lon pair
        :return: array of y indices, array of x indices, number of indices in transect
        """

        [j1, i1] = self.find_j_i(start[0], start[1])  # lat , lon
        [j2, i2] = self.find_j_i(end[0], end[1])  # lat , lon

        line_length = max(np.abs(j2 - j1), np.abs(i2 - i1)) + 1

        jj1 = [int(jj) for jj in np.round(np.linspace(j1, j2, num=line_length))]
        ii1 = [int(ii) for ii in np.round(np.linspace(i1, i2, num=line_length))]

        return jj1, ii1, line_length
    
    def nearest_xy_indices(self, model_dataset, new_lons, new_lats):
        '''
        Obtains the x and y indices of the nearest model points to specified
        lists of longitudes and latitudes. Makes use of sklearn.neighbours
        and its BallTree haversine method. 
        
        Example Useage
        ----------
        # Get indices of model points closest to altimetry points
        ind_x, ind_y = nemo.nearest_indices(altimetry.dataset.longitude,
                                            altimetry.dataset.latitude)
        # Nearest neighbour interpolation of model dataset to these points
        interpolated = nemo.dataset.isel(x_dim = ind_x, y_dim = ind_y)

        Parameters
        ----------
        model_dataset (xr.Dataset or xr.DataArray): model xarray dataset.
            Must contain coordinates.
        new_lons (array): Array of longitudes (degrees) to compare with model
        new_lats (array): Array of latitudes (degrees) to compare with model
        
        Returns
        -------
        Array of x indices, Array of y indices
        '''
        # Cast lat/lon to numpy arrays in case xarray things
        new_lons = np.array(new_lons)
        new_lats = np.array(new_lats)
        mod_lon = np.array(model_dataset.longitude).flatten()
        mod_lat = np.array(model_dataset.latitude).flatten()
        
        # Put lons and lats into 2D location arrays for BallTree: [lat, lon]
        mod_locs = np.vstack((mod_lat, mod_lon)).transpose()
        new_locs = np.vstack((new_lats, new_lons)).transpose()
        
        # Convert lat/lon to radians for BallTree
        mod_locs = np.radians(mod_locs)
        new_locs = np.radians(new_locs)
        
        # Do nearest neighbour interpolation using BallTree (gets indices)
        tree = nb.BallTree(mod_locs, leaf_size=5, metric='haversine')
        _, ind_1d = tree.query(new_locs, k=1)
        
        # Get 2D indices from 1D index output from BallTree
        ind_y, ind_x = np.unravel_index(ind_1d, model_dataset.longitude.shape)
        ind_x = xr.DataArray(ind_x.squeeze())
        ind_y = xr.DataArray(ind_y.squeeze())

        return ind_x, ind_y
    
    def nearest_time_indices(self):
        raise NotImplementedError
        return
    
    def nearest_depth_indices(self):
        raise NotImplementedError
        return
    
    def interpolate_in_space(self, model_array, new_lons, new_lats):
        '''
        Interpolates a provided xarray.DataArray in space to new longitudes
        and latitudes using a nearest neighbour method (BallTree).
        
        Example Useage
        ----------
        # Get an interpolated DataArray for temperature onto two locations
        interpolated = nemo.interpolate_in_space(nemo.dataset.votemper,
                                                 [0,1], [45,46])
        Parameters
        ----------
        model_array (xr.DataArray): Model variable DataArray to interpolate
        new_lons (1Darray): Array of longitudes (degrees) to compare with model
        new_lats (1Darray): Array of latitudes (degrees) to compare with model
        
        Returns
        -------
        Interpolated DataArray
        '''
        
        # Get nearest indices
        ind_x, ind_y = self.nearest_xy_indices(model_array, new_lons, new_lats)
        
        # Geographical interpolation (using BallTree indices)
        interpolated = model_array.isel(x_dim=ind_x, y_dim=ind_y)
        interpolated = interpolated.rename({'dim_0':'interp_dim'})
        return interpolated
    
    def interpolate_in_time(self, model_array, new_times, 
                               interp_method = 'nearest', extrapolate=True):
        '''
        Interpolates a provided xarray.DataArray in time to new python
        datetimes using a specified scipy.interpolate method.
        
        Example Useage
        ----------
        # Get an interpolated DataArray for temperature onto altimetry times
        new_times = altimetry.dataset.time
        interpolated = nemo.interpolate_in_space(nemo.dataset.votemper,
                                                 new_times)
        Parameters
        ----------
        model_array (xr.DataArray): Model variable DataArray to interpolate
        new_times (array): New times to interpolate to (array of datetimes)
        interp_method (str): Interpolation method
        
        Returns
        -------
        Interpolated DataArray
        '''
        
        # Time interpolation
        interpolated = model_array.swap_dims({'t_dim':'time'})
        if extrapolate:
            interpolated = interpolated.interp(time = new_times,
                                           method = interp_method,
                                           kwargs={'fill_value':'extrapolate'})
        else:
            interpolated = interpolated.interp(time = new_times,
                                           method = interp_method)
        #interpolated = interpolated.swap_dims({'time':'t_dim'})
        
        return interpolated
    
    def interpolate_in_depth(self, model_array, new_depths):
        raise NotImplementedError
        return
    
    def construct_density( self, EOS='EOS10' ):
        
        '''
            Constructs the in-situ density using the salinity, temperture and 
            depth_0 fields and adds a density attribute to the t-grid dataset 
            
            Requirements: The supplied t-grid dataset must contain the 
            Practical Salinity and the Potential Temperature variables. The depth_0
            field must also be supplied. The GSW package is used to calculate
            The Absolute Pressure, Absolute Salinity and Conservate Temperature.
            
            Note that currently density can only be constructed using the EOS10
            equation of state.

        Parameters
        ----------
        EOS : equation of state, optional
            DESCRIPTION. The default is 'EOS10'.


        Returns
        -------
        None.
        adds attribute NEMO.dataset.density

        '''  
        
        try:
            if EOS != 'EOS10': 
                raise ValueError(str(self) + ': Density calculation for ' + EOS + ' not implemented.')
            if self.grid_ref != 't-grid':
                raise ValueError(str(self) + ': Density calculation can only be performed for a t-grid object,\
                                 the tracer grid for NEMO.' )
        
            try:    
                shape_ds = ( self.dataset.t_dim.size, self.dataset.z_dim.size, 
                                self.dataset.y_dim.size, self.dataset.x_dim.size )
                sal = self.dataset.salinity.to_masked_array()
                temp = self.dataset.temperature.to_masked_array()                
            except AttributeError:
                shape_ds = ( 1, self.dataset.z_dim.size, 
                                self.dataset.y_dim.size, self.dataset.x_dim.size )
                sal = self.dataset.salinity.to_masked_array()[np.newaxis,...]
                temp = self.dataset.temperature.to_masked_array()[np.newaxis,...]
            
            density = np.ma.zeros( shape_ds )
            
            s_levels = self.dataset.depth_0.to_masked_array()
            lat = self.dataset.latitude.values
            lon = self.dataset.longitude.values
            # Absolute Pressure 
            pressure_absolute = np.ma.masked_invalid(
                gsw.p_from_z( -s_levels, lat ) ) # depth must be negative    
            # Absolute Salinity            
            sal_absolute = np.ma.masked_invalid(
                gsw.SA_from_SP( sal, pressure_absolute, lon, lat ) )
            sal_absolute = np.ma.masked_less(sal_absolute,0)
            # Conservative Temperature
            temp_conservative = np.ma.masked_invalid(
                gsw.CT_from_pt( sal_absolute, temp ) )
            # In-situ density
            density = np.ma.masked_invalid( gsw.rho( 
                sal_absolute, temp_conservative, pressure_absolute ) )
            
            coords={'depth_0': (('z_dim','y_dim','x_dim'), self.dataset.depth_0.values),
                    'latitude': (('y_dim','x_dim'), self.dataset.latitude.values),
                    'longitude': (('y_dim','x_dim'), self.dataset.longitude.values)}
            dims=['z_dim', 'y_dim', 'x_dim']
            attributes = {'units': 'kg / m^3', 'standard name': 'In-situ density'}
            
            if shape_ds[0] != 1:
                coords['time'] = (('t_dim'), self.dataset.time.values)
                dims.insert(0, 't_dim')
    
            self.dataset['density'] = xr.DataArray( np.squeeze(density), 
                    coords=coords, dims=dims, attrs=attributes )
            
        except AttributeError as err:
            print(err)
                                 
    

    def trim_domain_size( self, dataset_domain ):
        """
        Trim the domain variables if the dataset object is a spatial subset
        
        Note: This breaks if the SW & NW corner values of nav_lat and nav_lon 
        are masked, as can happen if on land...
        """
        if (self.dataset['x_dim'].size != dataset_domain['x_dim'].size)  \
                or (self.dataset['y_dim'].size != dataset_domain['y_dim'].size):
            #print('The domain  and dataset objects are different sizes:', \
            #      ' [{},{}] cf [{},{}]. Trim domain.'.format(
            #      dataset_domain['x_dim'].size, dataset_domain['y_dim'].size,
            #      self.dataset['x_dim'].size, self.dataset['y_dim'].size ))

            # Find the corners of the cut out domain.
            [j0,i0] = self.find_j_i_domain( self.dataset.nav_lat[0,0],
                                    self.dataset.nav_lon[0,0], dataset_domain )
            [j1,i1] = self.find_j_i_domain( self.dataset.nav_lat[-1,-1],
                                    self.dataset.nav_lon[-1,-1], dataset_domain )

            dataset_subdomain = dataset_domain.isel(
                                        y_dim = slice(j0, j1 + 1),
                                        x_dim = slice(i0, i1 + 1) )
            return dataset_subdomain
        else:
            return dataset_domain

    def copy_domain_vars_to_dataset( self, dataset_domain, grid_vars ):
        """
        Map the domain coordand metric variables to the dataset object.
        Expects the source and target DataArrays to be same sizes.
        """
        for var in grid_vars:
            try:
                new_name = self.var_mapping_domain[var]
                self.dataset[new_name] = dataset_domain[var].squeeze()
                #print("map: {} --> {}".format( var, new_name))
            except:
                pass

    def differentiate(self, in_varstr, dim='z_dim', out_varstr=None, out_obj=None):
        """
        Derivatives are computed in x_dim, y_dim, z_dim (or i,j,k) directions
        wrt lambda, phi, or z coordinates (with scale factor in metres not degrees).

        Derivatives are calculated using the approach adopted in NEMO,
        specifically using the 1st order accurate central difference
        approximation. For reference see section 3.1.2 (sec. Discrete operators)
        of the NEMO v4 Handbook.

        Currently the method does not accomodate all possible eventualities. It
        covers:
        1) d(grid_t)/dz --> grid_w
        
        Returns  an object (with the appropriate target grid_ref) containing
        derivative (out_varstr) as xr.DataArray
        
        This is hardwired to expect:
        1) depth_0 and e3_0 fields exist
        2) xr.DataArrays are 4D
        3) self.filename_domain if out_obj not specified
        4) If out_obj is not specified, one is built that is  the size of
            self.filename_domain. I.e. automatic subsetting of out_obj is not
            supported.
        
        Example usage:
        --------------
        # Initialise DataArrays
        nemo_t = coast.NEMO( fn_data, fn_domain, grid_ref='t-grid' )
        # Compute dT/dz
        nemo_w_1 = nemo_t.differentiate( 'temperature', dim='z_dim' )
        
        # For f(z)=-z. Compute df/dz = -1. Surface value is set to zero
        nemo_t.dataset['depth4D'],_ = xr.broadcast( nemo_t.dataset['depth_0'], nemo_t.dataset['temperature'] )
        nemo_w_4 = nemo_t.differentiate( 'depth4D', dim='z_dim', out_varstr='dzdz' )
        
        Provide an existing target NEMO object and target variable name:
        nemo_w_1 = nemo_t.differentiate( 'temperature', dim='z_dim', out_varstr='dTdz', out_obj=nemo_w_1 )
        
        
        Parameters
        ----------
        in_varstr : str, name of variable to differentiate
        dim : str, dimension to operate over. E.g. {'z_dim', 'y_dim', 'x_dim', 't_dim'}
        out_varstr : str, (optional) name of the target xr.DataArray
        out_obj : exiting NEMO obj to store xr.DataArray (optional)

        """
        #import xarray as xr

        new_units = ""

        # Check in_varstr exists in self.
        if hasattr( self.dataset, in_varstr ):
            # self.dataset[in_varstr] exists

            var = self.dataset[in_varstr] # for convenience

            nt = var.sizes['t_dim']
            nz = var.sizes['z_dim']
            ny = var.sizes['y_dim']
            nx = var.sizes['x_dim']

            ## Compute d(t_grid)/dz --> w-grid
            # Check grid_ref and dir. Determine target grid_ref.
            if (self.grid_ref == 't-grid') and (dim == 'z_dim'):
                out_grid = 'w-grid'

                # If out_obj exists check grid_ref, else create out_obj.
                if (out_obj is None) or (out_obj.grid_ref != out_grid):
                    try:
                        out_obj = NEMO( fn_domain=self.filename_domain, grid_ref=out_grid )
                    except:
                        print('Failed to create target NEMO obj. Perhaps self.',
                                'filename_domain={} is empty?'
                                .format(self.filename_domain))
                    #print('make new target object (out_obj)')

                # Check is out_varstr is defined, else create it
                if out_varstr is None:
                    out_varstr = in_varstr + '_dz'
                    #print('make new target variable name: out_str = {}'\
                    #      .format(out_varstr))

                # Create new DataArray with the same dimensions as the parent
                # Crucially have a coordinate value that is appropriate to the target location.
                blank = xr.zeros_like( var.isel(z_dim=[0]) ) # Using "z_dim=[0]" as a list preserves z-dimension
                blank.coords['depth_0'] -= blank.coords['depth_0'] # reset coord vals to zero
                # Add blank slice to the 'surface'. Concat over the 'dim' coords
                diff = xr.concat([blank, var.diff(dim)], dim)
                diff_ndim, e3w_ndim = xr.broadcast( diff, out_obj.dataset.e3_0.squeeze() )
                # Compute the derivative
                out_obj.dataset[out_varstr] = - diff_ndim / e3w_ndim

                # Assign attributes
                new_units = var.units+'/'+ out_obj.dataset.depth_0.units
                # Convert to a xr.DataArray and return
                out_obj.dataset[out_varstr].attrs = {
                                           'units': new_units,
                                           'standard_name': out_varstr}

                # Return in object.
                return out_obj

            else:
                print('Not ready for that combination of grid ({}) and ',\
                'derivative ({})'.format(self.grid_ref, dim))
                return None
        else:
            print('{} does not exist in self.dataset'.format(in_varstr))
            return None
        
        
